import os.path
import enum
import datetime

from sqlalchemy import (
    Column,
    BigInteger,
    Boolean,
    Integer,
    Interval,
    String,
    Numeric,
    DateTime,
    JSON,
    Float,
    Text,
    ForeignKey,
    DDL,
    event,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.types import ARRAY, Enum
from sqlalchemy.sql import func

# One declarative_base object for each schema created
Base = declarative_base()

schemas = (
    "CREATE SCHEMA IF NOT EXISTS triage_metadata;"
    " CREATE SCHEMA IF NOT EXISTS test_results;"
    " CREATE SCHEMA IF NOT EXISTS train_results;"
)

event.listen(Base.metadata, "before_create", DDL(schemas))

group_proc_filename = os.path.join(
    os.path.dirname(__file__), "sql", "model_group_stored_procedure.sql"
)
with open(group_proc_filename) as fd:
    stmt = fd.read()

event.listen(Base.metadata, "before_create", DDL(stmt))

nuke_triage_filename = os.path.join(
    os.path.dirname(__file__), "sql", "nuke_triage.sql"
)
with open(nuke_triage_filename) as fd:
    stmt = fd.read()


event.listen(Base.metadata, "before_create", DDL(stmt.replace('%', '%%')))


class Experiment(Base):

    __tablename__ = "experiments"
    __table_args__ = {"schema": "triage_metadata"}

    experiment_hash = Column(String, primary_key=True)
    config = Column(JSONB)
    time_splits = Column(Integer)
    as_of_times = Column(Integer)
    feature_blocks = Column(Integer)
    total_features = Column(Integer)
    feature_group_combinations = Column(Integer)
    matrices_needed = Column(Integer)
    grid_size = Column(Integer)
    models_needed = Column(Integer)
    random_seed = Column(Integer)


class ExperimentRunStatus(enum.Enum):
    started = 1
    completed = 2
    failed = 3


class ExperimentRun(Base):

    __tablename__ = "experiment_runs"
    __table_args__ = {"schema": "triage_metadata"}

    run_id = Column("id", Integer, primary_key=True)
    start_time = Column(DateTime)
    start_method = Column(String)
    git_hash = Column(String)
    triage_version = Column(String)
    python_version = Column(String)
    experiment_hash = Column(
        String,
        ForeignKey("triage_metadata.experiments.experiment_hash")
    )
    platform = Column(Text)
    os_user = Column(Text)
    working_directory = Column(Text)
    ec2_instance_type = Column(Text)
    log_location = Column(Text)
    experiment_class_path = Column(Text)
    experiment_kwargs = Column(JSONB)
    installed_libraries = Column(ARRAY(Text))
    matrix_building_started = Column(DateTime)
    matrices_made = Column(Integer, default=0)
    matrices_skipped = Column(Integer, default=0)
    matrices_errored = Column(Integer, default=0)
    model_building_started = Column(DateTime)
    models_made = Column(Integer, default=0)
    models_skipped = Column(Integer, default=0)
    models_errored = Column(Integer, default=0)
    last_updated_time = Column(DateTime, onupdate=datetime.datetime.now)
    current_status = Column(Enum(ExperimentRunStatus))
    stacktrace = Column(Text)
    random_seed = Column(Integer)
    experiment_rel = relationship("Experiment")


class Subset(Base):

    __tablename__ = "subsets"
    __table_args__ = {"schema": "triage_metadata"}

    subset_hash = Column(String, primary_key=True)
    config = Column(JSONB)
    created_timestamp = Column(DateTime(timezone=True), server_default=func.now())


class ModelGroup(Base):

    __tablename__ = "model_groups"
    __table_args__ = {"schema": "triage_metadata"}

    model_group_id = Column(Integer, primary_key=True)
    model_type = Column(Text)
    hyperparameters = Column(JSONB)
    feature_list = Column(ARRAY(Text))
    model_config = Column(JSONB)


class ListPrediction(Base):

    __tablename__ = "list_predictions"
    __table_args__ = {"schema": "triage_metadata"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    entity_id = Column(BigInteger, primary_key=True)
    as_of_date = Column(DateTime, primary_key=True)
    score = Column(Numeric)
    rank_abs = Column(Integer)
    rank_pct = Column(Float)
    matrix_uuid = Column(Text)
    test_label_timespan = Column(Interval)

    model_rel = relationship("Model")


class ExperimentMatrix(Base):
    __tablename__ = "experiment_matrices"
    __table_args__ = {"schema": "triage_metadata"}

    experiment_hash = Column(
        String,
        ForeignKey("triage_metadata.experiments.experiment_hash"),
        primary_key=True
    )

    matrix_uuid = Column(String, primary_key=True)


class Matrix(Base):

    __tablename__ = "matrices"
    __table_args__ = {"schema": "triage_metadata"}

    matrix_id = Column(String)
    matrix_uuid = Column(String, unique=True, index=True, primary_key=True)
    matrix_type = Column(String)  # 'train' or 'test'
    labeling_window = Column(Interval)
    num_observations = Column(Integer)
    creation_time = Column(DateTime(timezone=True), server_default=func.now())
    lookback_duration = Column(Interval)
    feature_start_time = Column(DateTime)
    matrix_metadata = Column(JSONB)
    built_by_experiment = Column(
        String, ForeignKey("triage_metadata.experiments.experiment_hash")
    )
    feature_dictionary = Column(JSONB)


class Model(Base):

    __tablename__ = "models"
    __table_args__ = {"schema": "triage_metadata"}

    model_id = Column(Integer, primary_key=True)
    model_group_id = Column(
        Integer, ForeignKey("triage_metadata.model_groups.model_group_id")
    )
    model_hash = Column(String, unique=True, index=True)
    run_time = Column(DateTime)
    batch_run_time = Column(DateTime)
    model_type = Column(String)
    hyperparameters = Column(JSONB)
    model_comment = Column(Text)
    batch_comment = Column(Text)
    config = Column(JSON)
    built_by_experiment = Column(
        String, ForeignKey("triage_metadata.experiments.experiment_hash")
    )
    built_in_experiment_run = Column(
        Integer, ForeignKey("triage_metadata.experiment_runs.id")
    )
    train_end_time = Column(DateTime)
    test = Column(Boolean)
    train_matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"))
    training_label_timespan = Column(Interval)
    model_size = Column(Float)
    random_seed = Column(Integer)

    model_group_rel = relationship("ModelGroup")
    matrix_rel = relationship("Matrix")

    def delete(self, session):
        # basically implement a cascade, in case cascade is not implemented
        (session.query(FeatureImportance).filter_by(model_id=self.model_id).delete())
        (session.query(TestEvaluation).filter_by(model_id=self.model_id).delete())
        (session.query(TestPrediction).filter_by(model_id=self.model_id).delete())
        session.delete(self)


class ExperimentModel(Base):
    __tablename__ = "experiment_models"
    __table_args__ = {"schema": "triage_metadata"}

    experiment_hash = Column(
        String,
        ForeignKey("triage_metadata.experiments.experiment_hash"),
        primary_key=True
    )
    model_hash = Column(String, primary_key=True)

    model_rel = relationship("Model", primaryjoin=(Model.model_hash == model_hash), foreign_keys=model_hash)
    experiment_rel = relationship("Experiment")


class FeatureImportance(Base):

    __tablename__ = "feature_importances"
    __table_args__ = {"schema": "train_results"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    model = relationship(Model)
    feature = Column(String, primary_key=True)
    feature_importance = Column(Numeric)
    rank_abs = Column(Integer)
    rank_pct = Column(Float)

    model_rel = relationship("Model")


class TestPrediction(Base):

    __tablename__ = "predictions"
    __table_args__ = {"schema": "test_results"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    entity_id = Column(BigInteger, primary_key=True)
    as_of_date = Column(DateTime, primary_key=True)
    score = Column(Numeric(6, 5))
    label_value = Column(Integer)
    rank_abs_no_ties = Column(Integer)
    rank_abs_with_ties = Column(Integer)
    rank_pct_no_ties = Column(Numeric(6, 5))
    rank_pct_with_ties = Column(Numeric(6, 5))
    matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"))
    test_label_timespan = Column(Interval)

    model_rel = relationship("Model")
    matrix_rel = relationship("Matrix")


class TrainPrediction(Base):

    __tablename__ = "predictions"
    __table_args__ = {"schema": "train_results"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    entity_id = Column(BigInteger, primary_key=True)
    as_of_date = Column(DateTime, primary_key=True)
    score = Column(Numeric(6, 5))
    label_value = Column(Integer)
    rank_abs_no_ties = Column(Integer)
    rank_abs_with_ties = Column(Integer)
    rank_pct_no_ties = Column(Float)
    rank_pct_with_ties = Column(Float)
    matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"))
    test_label_timespan = Column(Interval)

    model_rel = relationship("Model")
    matrix_rel = relationship("Matrix")


class TestPredictionMetadata(Base):
    __tablename__ = "prediction_metadata"
    __table_args__ = {"schema": "test_results"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"), primary_key=True)
    tiebreaker_ordering = Column(Text)
    random_seed = Column(Integer)
    predictions_saved = Column(Boolean)


class TrainPredictionMetadata(Base):
    __tablename__ = "prediction_metadata"
    __table_args__ = {"schema": "train_results"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"), primary_key=True)
    tiebreaker_ordering = Column(Text)
    random_seed = Column(Integer)
    predictions_saved = Column(Boolean)

class IndividualImportance(Base):

    __tablename__ = "individual_importances"
    __table_args__ = {"schema": "test_results"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    entity_id = Column(BigInteger, primary_key=True)
    as_of_date = Column(DateTime, primary_key=True)
    feature = Column(String, primary_key=True)
    method = Column(String, primary_key=True)
    feature_value = Column(Float)
    importance_score = Column(Float)

    model_rel = relationship("Model")


class TestEvaluation(Base):

    __tablename__ = "evaluations"
    __table_args__ = {"schema": "test_results"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    subset_hash = Column(String, primary_key=True, default='')
    evaluation_start_time = Column(DateTime, primary_key=True)
    evaluation_end_time = Column(DateTime, primary_key=True)
    as_of_date_frequency = Column(Interval, primary_key=True)
    matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"))
    metric = Column(String, primary_key=True)
    parameter = Column(String, primary_key=True)
    num_labeled_examples = Column(Integer)
    num_labeled_above_threshold = Column(Integer)
    num_positive_labels = Column(Integer)
    sort_seed = Column(Integer)
    best_value = Column(Numeric)
    worst_value = Column(Numeric)
    stochastic_value = Column(Numeric)
    num_sort_trials = Column(Integer)
    standard_deviation = Column(Numeric)

    matrix_rel = relationship("Matrix")
    model_rel = relationship("Model")


class TrainEvaluation(Base):

    __tablename__ = "evaluations"
    __table_args__ = {"schema": "train_results"}

    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True
    )
    subset_hash = Column(String, primary_key=True, default='')
    evaluation_start_time = Column(DateTime, primary_key=True)
    evaluation_end_time = Column(DateTime, primary_key=True)
    as_of_date_frequency = Column(Interval, primary_key=True)
    matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"))
    metric = Column(String, primary_key=True)
    parameter = Column(String, primary_key=True)
    num_labeled_examples = Column(Integer)
    num_labeled_above_threshold = Column(Integer)
    num_positive_labels = Column(Integer)
    sort_seed = Column(Integer)
    best_value = Column(Numeric)
    worst_value = Column(Numeric)
    stochastic_value = Column(Numeric)
    num_sort_trials = Column(Integer)
    standard_deviation = Column(Numeric)

    matrix_rel = relationship("Matrix")
    model_rel = relationship("Model")


class TestAequitas(Base):
    __tablename__ = "aequitas"
    __table_args__ = {"schema": "test_results"}
    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True, index=True
    )
    subset_hash = Column(String, primary_key=True, default='')
    tie_breaker = Column(String, primary_key=True)
    evaluation_start_time = Column(DateTime, primary_key=True, index=True)
    evaluation_end_time = Column(DateTime, primary_key=True, index=True)
    matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"))
    parameter = Column(String, primary_key=True, index=True)
    attribute_name = Column(String, primary_key=True, index=True)
    attribute_value = Column(String, primary_key=True, index=True)
    total_entities = Column(Integer)
    group_label_pos = Column(Integer)
    group_label_neg = Column(Integer)
    group_size = Column(Integer)
    group_size_pct = Column(Numeric)
    prev = Column(Numeric)
    pp = Column(Integer)
    pn = Column(Integer)
    fp = Column(Integer)
    fn = Column(Integer)
    tn = Column(Integer)
    tp = Column(Integer)
    ppr = Column(Numeric)
    pprev = Column(Numeric)
    tpr = Column(Numeric)
    tnr = Column(Numeric)
    for_ = Column("for", Numeric)
    fdr = Column(Numeric)
    fpr = Column(Numeric)
    fnr = Column(Numeric)
    npv = Column(Numeric)
    precision = Column(Numeric)
    ppr_disparity = Column(Numeric)
    ppr_ref_group_value = Column(String)
    pprev_disparity = Column(Numeric)
    pprev_ref_group_value = Column(String)
    precision_disparity = Column(Numeric)
    precision_ref_group_value = Column(String)
    fdr_disparity = Column(Numeric)
    fdr_ref_group_value = Column(String)
    for_disparity = Column(Numeric)
    for_ref_group_value = Column(String)
    fpr_disparity = Column(Numeric)
    fpr_ref_group_value = Column(String)
    fnr_disparity = Column(Numeric)
    fnr_ref_group_value = Column(String)
    tpr_disparity = Column(Numeric)
    tpr_ref_group_value = Column(String)
    tnr_disparity = Column(Numeric)
    tnr_ref_group_value = Column(String)
    npv_disparity = Column(Numeric)
    npv_ref_group_value = Column(String)
    Statistical_Parity = Column(Boolean)
    Impact_Parity = Column(Boolean)
    FDR_Parity = Column(Boolean)
    FPR_Parity = Column(Boolean)
    FOR_Parity = Column(Boolean)
    FNR_Parity = Column(Boolean)
    TypeI_Parity = Column(Boolean)
    TypeII_Parity = Column(Boolean)
    Equalized_Odds = Column(Boolean)
    Unsupervised_Fairness = Column(Boolean)
    Supervised_Fairness = Column(Boolean)

    matrix_rel = relationship("Matrix")
    model_rel = relationship("Model")


class TrainAequitas(Base):
    __tablename__ = "aequitas"
    __table_args__ = {"schema": "train_results"}
    model_id = Column(
        Integer, ForeignKey("triage_metadata.models.model_id"), primary_key=True, index=True
    )
    subset_hash = Column(String, primary_key=True, default='')
    tie_breaker = Column(String, primary_key=True)
    evaluation_start_time = Column(DateTime, primary_key=True, index=True)
    evaluation_end_time = Column(DateTime, primary_key=True, index=True)
    matrix_uuid = Column(Text, ForeignKey("triage_metadata.matrices.matrix_uuid"))
    parameter = Column(String, primary_key=True, index=True)
    attribute_name = Column(String, primary_key=True, index=True)
    attribute_value = Column(String, primary_key=True, index=True)
    total_entities = Column(Integer)
    group_label_pos = Column(Integer)
    group_label_neg = Column(Integer)
    group_size = Column(Integer)
    group_size_pct = Column(Numeric)
    prev = Column(Numeric)
    pp = Column(Integer)
    pn = Column(Integer)
    fp = Column(Integer)
    fn = Column(Integer)
    tn = Column(Integer)
    tp = Column(Integer)
    ppr = Column(Numeric)
    pprev = Column(Numeric)
    tpr = Column(Numeric)
    tnr = Column(Numeric)
    for_ = Column("for", Numeric)
    fdr = Column(Numeric)
    fpr = Column(Numeric)
    fnr = Column(Numeric)
    npv = Column(Numeric)
    precision = Column(Numeric)
    ppr_disparity = Column(Numeric)
    ppr_ref_group_value = Column(String)
    pprev_disparity = Column(Numeric)
    pprev_ref_group_value = Column(String)
    precision_disparity = Column(Numeric)
    precision_ref_group_value = Column(String)
    fdr_disparity = Column(Numeric)
    fdr_ref_group_value = Column(String)
    for_disparity = Column(Numeric)
    for_ref_group_value = Column(String)
    fpr_disparity = Column(Numeric)
    fpr_ref_group_value = Column(String)
    fnr_disparity = Column(Numeric)
    fnr_ref_group_value = Column(String)
    tpr_disparity = Column(Numeric)
    tpr_ref_group_value = Column(String)
    tnr_disparity = Column(Numeric)
    tnr_ref_group_value = Column(String)
    npv_disparity = Column(Numeric)
    npv_ref_group_value = Column(String)
    Statistical_Parity = Column(Boolean)
    Impact_Parity = Column(Boolean)
    FDR_Parity = Column(Boolean)
    FPR_Parity = Column(Boolean)
    FOR_Parity = Column(Boolean)
    FNR_Parity = Column(Boolean)
    TypeI_Parity = Column(Boolean)
    TypeII_Parity = Column(Boolean)
    Equalized_Odds = Column(Boolean)
    Unsupervised_Fairness = Column(Boolean)
    Supervised_Fairness = Column(Boolean)

    matrix_rel = relationship("Matrix")
    model_rel = relationship("Model")


hash_partitioning_filename = os.path.join(
    os.path.dirname(__file__), "sql", "predictions_hash_partitioning.sql"
)
with open(hash_partitioning_filename) as fd:
    stmt = fd.read()

event.listen(Base.metadata, "after_create", DDL(stmt))



from triage.component.postmodeling.contrast.model_group_evaluator import ModelGroupEvaluator
from triage.component.postmodeling.contrast.model_evaluator import ModelEvaluator
import pytest


@pytest.fixture(scope="module")
def model_group_evaluator(finished_experiment_without_predictions):
    return ModelGroupEvaluator((1, 1), finished_experiment_without_predictions.db_engine)


@pytest.fixture(scope="module")
def model_evaluator(finished_experiment_without_predictions):
    return ModelEvaluator(1, 1, finished_experiment_without_predictions.db_engine)


def test_ModelGroupEvaluator_metadata(model_group_evaluator):
    assert all(value for metadata_row in model_group_evaluator.metadata for key, value in metadata_row.items() )


def test_ModelGroupEvaluator_predictions(model_group_evaluator):
    with pytest.raises(RuntimeError):
        model_group_evaluator.predictions


def test_ModelEvaluator_metadata(model_evaluator):
    assert all(value for key, value in model_evaluator.metadata.items())


def test_ModelEvaluator_predictions(model_evaluator):
    with pytest.raises(RuntimeError):
        model_evaluator.predictions



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import traceback
from functools import partial
from pebble import ProcessPool
from multiprocessing.reduction import ForkingPickler

from triage.component.catwalk.utils import Batch

from triage.experiments import ExperimentBase


class MultiCoreExperiment(ExperimentBase):
    def __init__(self, config, db_engine, *args, n_processes=1, n_db_processes=1, **kwargs):
        try:
            ForkingPickler.dumps(db_engine)
        except Exception as exc:
            raise TypeError(
                "multiprocessing is unable to pickle passed SQLAlchemy engine. "
                "use triage.create_engine instead when running MultiCoreExperiment: "
                "(e.g. from triage import create_engine)"
            ) from exc

        super(MultiCoreExperiment, self).__init__(config, db_engine, *args, **kwargs)
        if n_processes < 1:
            raise ValueError("n_processes must be 1 or greater")
        if n_db_processes < 1:
            raise ValueError("n_db_processes must be 1 or greater")
        if n_db_processes == 1 and n_processes == 1:
            logger.notice(
                "Both n_processes and n_db_processes were set to 1. "
                "If you only wish to use one process to run the experiment, "
                "consider using the SingleThreadedExperiment class instead"
            )
        self.n_processes = n_processes
        self.n_db_processes = n_db_processes

    def generated_chunked_parallelized_results(
        self, partially_bound_function, tasks, n_processes, chunksize=1
    ):
        with ProcessPool(n_processes, max_tasks=1) as pool:
            future = pool.map(
                partially_bound_function,
                [list(task_batch) for task_batch in Batch(tasks, chunksize)],
            )
            iterator = future.result()
            while True:
                try:
                    yield next(iterator)
                except StopIteration:
                    break
                except Exception:
                    logger.exception('Child failure')

    def process_train_test_batches(self, batches):
        partial_test = partial(
            run_task_with_splatted_arguments, self.model_train_tester.process_task
        )

        for batch in batches:
            if batch.parallelizable:
                logger.info(
                    f"Starting parallelizable batch train/testing with {len(batch.tasks)} tasks, {self.n_processes} processes",
                )
                parallelize(partial_test, batch.tasks, self.n_processes)
            else:
                logger.info(
                    "Starting serial batch train/testing with {len(batch.tasks)} tasks",
                )
                for serial_task in batch.tasks:
                    self.model_train_tester.process_task(**serial_task)

    def process_query_tasks(self, query_tasks):
        logger.info("Processing query tasks with %s processes", self.n_db_processes)
        for table_name, tasks in query_tasks.items():
            logger.info("Processing features for %s", table_name)
            self.feature_generator.run_commands(tasks.get("prepare", []))
            partial_insert = partial(
                insert_into_table, feature_generator=self.feature_generator
            )

            insert_batches = [
                list(task_batch) for task_batch in Batch(tasks.get("inserts", []), 25)
            ]
            parallelize(partial_insert, insert_batches, n_processes=self.n_db_processes)
            self.feature_generator.run_commands(tasks.get("finalize", []))
            logger.info(f"{table_name} completed")

    def process_matrix_build_tasks(self, matrix_build_tasks):
        partial_build_matrix = partial(
            run_task_with_splatted_arguments, self.matrix_builder.build_matrix
        )
        logger.info(
            f"Starting parallel matrix building: {len(self.matrix_build_tasks.keys())} matrices, {self.n_processes} processes",
        )
        parallelize(
            partial_build_matrix, self.matrix_build_tasks.values(), self.n_processes
        )

    def process_subset_tasks(self, subset_tasks):
        partial_subset = partial(
            run_task_with_splatted_arguments, self.subsetter.process_task
        )

        logger.info(
            f"Starting parallel subset creation: {len(subset_tasks)} subsets, {self.n_db_processes} processes",
        )
        parallelize(
            partial_subset, subset_tasks, self.n_db_processes
        )


def insert_into_table(insert_statements, feature_generator):
    try:
        logger.info("Beginning insert batch")
        feature_generator.run_commands(insert_statements)
        return True
    except Exception:
        logger.exception("Child error")
        return False


def parallelize(partially_bound_function, tasks, n_processes):
    num_successes = 0
    num_failures = 0
    results = []
    with ProcessPool(n_processes, max_tasks=1) as pool:
        future = pool.map(partially_bound_function, tasks)
        iterator = future.result()
        results = []
        while True:
            try:
                result = next(iterator)
            except StopIteration:
                break
            except Exception:
                logger.exception('Child failure')
                num_failures += 1
            else:
                results.append(result)
                num_successes += 1

        logger.info("Done. successes: %s, failures: %s", num_successes, num_failures)
        return results


def run_task_with_splatted_arguments(task_runner, task):
    try:
        return task_runner(**task)
    except Exception:
        logger.exception("Child error")
        return None



"""Main application"""
from .model_trainers import ModelTrainer
from .predictors import Predictor
from .evaluation import ModelEvaluator
from .individual_importance import IndividualImportanceCalculator, IndividualImportanceCalculatorNoOp
from .model_grouping import ModelGrouper
from .subsetters import Subsetter, SubsetterNoOp
from .protected_groups_generators import ProtectedGroupsGenerator, ProtectedGroupsGeneratorNoOp
from .utils import filename_friendly_hash

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from collections import namedtuple

import numpy as np

TaskBatch = namedtuple('TaskBatch', ['parallelizable', 'tasks', 'description'])


class ModelTrainTester:
    def __init__(
        self,
        matrix_storage_engine,
        model_trainer,
        model_evaluator,
        individual_importance_calculator,
        predictor,
        subsets,
        protected_groups_generator,
        cohort_hash=None,
        replace=True
    ):
        self.matrix_storage_engine = matrix_storage_engine
        self.model_trainer = model_trainer
        self.model_evaluator = model_evaluator
        self.individual_importance_calculator = individual_importance_calculator
        self.predictor = predictor
        self.subsets = subsets
        self.replace = replace
        self.protected_groups_generator = protected_groups_generator
        self.cohort_hash = cohort_hash

    def generate_task_batches(self, splits, grid_config, model_comment=None):
        train_test_tasks = []
        logger.debug(f"Generating train/test tasks for {len(splits)} splits")
        for split in splits:
            train_store = self.matrix_storage_engine.get_store(split["train_uuid"])
            train_tasks = self.model_trainer.generate_train_tasks(
                grid_config=grid_config,
                misc_db_parameters=dict(test=False, model_comment=model_comment),
                matrix_store=train_store
            )

            for test_matrix_def, test_uuid in zip(
                split["test_matrices"], split["test_uuids"]
            ):
                test_store = self.matrix_storage_engine.get_store(test_uuid)

                for train_task in train_tasks:
                    train_test_tasks.append(
                        {
                            "test_store": test_store,
                            "train_store": train_store,
                            "train_kwargs": train_task,
                        }
                    )
        return self.order_and_batch_tasks(train_test_tasks)


    def order_and_batch_tasks(self, tasks):
        batches = (
            TaskBatch(
                parallelizable=True,
                tasks=[],
                description="Baselines or simple classifiers (e.g. DecisionTree, SLR)"
            ),
            TaskBatch(
                parallelizable=False,
                tasks=[],
                description="Heavyweight classifiers with n_jobs set to -1."
            ),
            TaskBatch(
                parallelizable=True,
                tasks=[],
                description="All classifiers not found in one of the other batches (e.g. gradient boosting)."
            ),
        )

        for task in tasks:
            if task['train_kwargs']['class_path'].startswith('triage.component.catwalk.baselines') \
                    or task['train_kwargs']['class_path'] in (
                    'triage.component.catwalk.estimators.classifiers.ScaledLogisticRegression',
                    'sklearn.tree.DecisionTreeClassifier',
                    'sklearn.dummy.DummyClassifier'
                    ):
                # First priority: baselines or simple, effective classifiers
                batches[0].tasks.append(task)
            elif task['train_kwargs']['parameters'].get('n_jobs', None) == -1:
                # Second priority: heavyweight classifiers that we use the whole machines for
                batches[1].tasks.append(task)
            else:
                # Last priority: Everything else. Maybe these are slow/non-parallelizable
                batches[2].tasks.append(task)
        logger.verbose("Split train/test tasks into three task batches. - each batch has models from all splits")
        for batch_num, batch in enumerate(batches, 1):
            logger.verbose(f"Batch {batch_num}: {batch.description} ({len(batch.tasks)} tasks total)")



        return batches


    def process_all_batches(self, task_batches):
        for n_batch, batch in enumerate(task_batches, start=1):
            logger.verbose(f"Processing '{batch.description}' [{n_batch} of {len(task_batches)} batches]")
            for n_task, task in enumerate(batch.tasks, start=1):
                logger.verbose(f"Processing task [{n_task} of {len(batch.tasks)}] from {batch.description}")
                self.process_task(**task)
                logger.verbose(f"Task {n_task} from {batch.description} completed")
            logger.success(f"Batch '{batch.description}' completed")

    def process_task(self, test_store, train_store, train_kwargs):
        logger.verbose(f"Training {train_kwargs.get('class_path')}({train_kwargs.get('parameters')}) [{train_kwargs.get('model_hash')}] on train matrix {train_store.uuid}")

        # If the matrices and train labels are OK, train and test the model!
        with self.model_trainer.cache_models(), test_store.cache(), train_store.cache():
            # will cache any trained models until it goes out of scope (at the end of the task)
            # this way we avoid loading the model pickle again for predictions

            # If the train or test design matrix empty, or if the train store only
            # has one label value, skip training the model.
            if train_store.empty:
                logger.notice(
                    f"""Train matrix for split {train_store.uuid} was empty,
                    no point in training this model. Skipping
                    """
                )
                return

            if len(train_store.labels.unique()) == 1:
                logger.notice(
                    f"""Train Matrix for split {train_store.uuid} had only one
                    unique value, no point in training this model. Skipping
                    """
                )
                return

            if test_store.empty:
                logger.notice(
                    f"""Test matrix for uuid {test_store.uuid}
                    was empty, no point in generating predictions. Not processing train/test task.
                    """
                )
                return

            model_id = self.model_trainer.process_train_task(**train_kwargs)

            if not model_id:
                logger.warning("Training unsuccessful for {train_kwargs.get('class_path')}({train_kwargs.get('parameters')}) [{train_kwargs.get('model_hash')}] on train matrix {train_store.uuid}. "
                               "No model id returned.  Not attempting to test it")
                return

            logger.success(f"Trained model id {model_id}: {train_kwargs.get('class_path')}({train_kwargs.get('parameters')}) [{train_kwargs.get('model_hash')}] on train matrix {train_store.uuid}. ")

            # Storing individual importances (if any)
            self.individual_importance_calculator.calculate_and_save_all_methods_and_dates(
                model_id, test_store
            )

            as_of_dates = test_store.as_of_dates
            logger.debug(
                f"Testing and evaluating model {model_id}  {train_kwargs.get('class_path')}({train_kwargs.get('parameters')}) [{train_kwargs.get('model_hash')}] "
                f"on test matrix {test_store.uuid}. ")
            logger.spam(f"as_of_times min: {min(as_of_dates)} max: {max(as_of_dates)} num: {len(as_of_dates)}")


            # Generate predictions for the testing data then training data
            for store in (test_store, train_store):
                if self.replace or self.model_evaluator.needs_evaluations(store, model_id):
                    logger.spam(
                        f"Generating new predictions for "
                        f"{store.matrix_type.string_name} matrix {store.uuid}, and model {model_id} to make evaluation",
                    )

                    predictions_proba = self.predictor.predict(
                        model_id,
                        store,
                        misc_db_parameters=dict(),
                        train_matrix_columns=train_store.columns(),
                    )

                    logger.debug(f"Predictions generated for {store.matrix_type.string_name} matrix {store.uuid} using model {model_id}")


                    protected_df = self.protected_groups_generator.as_dataframe(
                        as_of_dates=store.as_of_dates,
                        cohort_hash=self.cohort_hash,
                    )

                    logger.spam(
                        f"Evaluating model {model_id} on {store.matrix_type.string_name} matrix {store.uuid} "
                    )

                    self.model_evaluator.evaluate(
                        predictions_proba=predictions_proba,
                        matrix_store=store,
                        model_id=model_id,
                        subset=None,
                        protected_df=protected_df
                    )

                    logger.info(
                        f"Model {model_id} evaluation on {store.matrix_type.string_name} matrix {store.uuid} completed."
                    )

                else:
                    logger.notice(
                        f"The evaluations needed for {store.matrix_type.string_name} matrix {store.uuid} and "
                        f"model {model_id} are all present"
                        f"in db from a previous run (or none needed at all), so skipping!",
                    )


                for subset in self.subsets:
                    subset_hash = filename_friendly_hash(subset)
                    if self.replace or self.model_evaluator.needs_evaluations(store, model_id, subset_hash):

                        logger.spam(
                            f"Evaluating {store.matrix_type.string_name} matrix {store.uuid}, subset {subset_hash}, and model {model_id}"
                        )


                        self.model_evaluator.evaluate(
                            predictions_proba=predictions_proba,
                            matrix_store=store,
                            model_id=model_id,
                            subset=subset,
                            protected_df=protected_df
                        )

                        logger.info(
                            f"Model {model_id} evaluation on subset {filename_friendly_hash(subset)} of {store.matrix_type.string_name} matrix {store.uuid} completed."
                        )

                    else:
                        logger.notice(
                            f"The evaluations needed for {store.matrix_type.string_name} matrix {store.uuid}, subset {filename_friendly_hash(subset)}, and "
                            f"model {model_id} are all present"
                            f"in db from a previous run (or none needed at all), so skipping!",
                        )


__all__ = (
    "IndividualImportanceCalculator",
    "ModelEvaluator",
    "ModelGrouper"
    "ModelTrainer",
    "Predictor",
    "ModelTrainTester",
    "Subsetter",
)



import matplotlib
matplotlib.use('Agg')
import pytest
import testing.postgresql
import tempfile
from tests.utils import sample_config, populate_source_data
from triage import create_engine
from triage.component.catwalk.storage import ProjectStorage
from triage.component.catwalk.db import ensure_db
from tests.results_tests.factories import init_engine
from triage.component.postmodeling.crosstabs import CrosstabsConfigLoader
from triage.experiments import SingleThreadedExperiment


@pytest.fixture(name='db_engine', scope='function')
def fixture_db_engine():
    """pytest fixture provider to set up and teardown a "test" database
    and provide the test function a connection engine with which to
    query that database.

    """
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        yield engine
        engine.dispose()


@pytest.fixture(scope="function")
def db_engine_with_results_schema(db_engine):
    ensure_db(db_engine)
    init_engine(db_engine)
    yield db_engine


@pytest.fixture(scope="function")
def project_path():
    with tempfile.TemporaryDirectory() as temp_dir:
        yield temp_dir


@pytest.fixture(scope="function")
def project_storage(project_path):
    """Set up a temporary project storage engine on the filesystem

    Yields (catwalk.storage.ProjectStorage)
    """
    yield ProjectStorage(project_path)


@pytest.fixture(scope='module')
def shared_db_engine():
    """pytest fixture provider to set up and teardown a "test" database
    and provide a test module a connection engine with which to
    query that database.

    """
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        yield engine
        engine.dispose()


@pytest.fixture(scope="module")
def shared_project_storage():
    """Set up a temporary project storage engine on the filesystem at module scope

    Yields (catwalk.storage.ProjectStorage)
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        project_storage = ProjectStorage(temp_dir)
        yield project_storage


@pytest.fixture(scope="module")
def finished_experiment(shared_db_engine, shared_project_storage):
    """A successfully-run experiment. Its database schemas and project storage can be queried.

    Returns: (triage.experiments.SingleThreadedExperiment)
    """
    populate_source_data(shared_db_engine)
    base_config = sample_config()
    experiment = SingleThreadedExperiment(
        base_config,
        db_engine=shared_db_engine,
        project_path=shared_project_storage.project_path
    )
    experiment.run()
    return experiment


@pytest.fixture(scope="module")
def finished_experiment_without_predictions(shared_db_engine, shared_project_storage):
    """A successfully-run experiment. Its database schemas and project storage can be queried.

    Returns: (triage.experiments.SingleThreadedExperiment)
    """
    populate_source_data(shared_db_engine)
    base_config = sample_config()
    experiment = SingleThreadedExperiment(
        base_config,
        db_engine=shared_db_engine,
        project_path=shared_project_storage.project_path,
        save_predictions=False
    )
    experiment.run()
    return experiment


@pytest.fixture(scope='module')
def crosstabs_config():
    """Example crosstabs config.

    Should work with after an experiment run with tests.utils.sample_config
    """
    return CrosstabsConfigLoader(config={
        "output": {
          "schema": 'test_results',
          "table": 'crosstabs'
        },
        "thresholds": {
            "rank_abs": [50],
            "rank_pct": [],
        },
        "entity_id_list": [],
        "models_list_query": "select unnest(ARRAY[1]) :: int as model_id",
        "as_of_dates_query": "select unnest(ARRAY['2012-06-01']) :: date as as_of_date",
        "models_dates_join_query": """
    select model_id,
          as_of_date
          from models_list_query m
          cross join as_of_dates_query a join (select distinct model_id, as_of_date from test_results.predictions) p
          using (model_id, as_of_date)""",
        "features_query": """
select m.model_id, f1.*
 from features.entity_features_aggregation_imputed f1 join
 models_dates_join_query m using (as_of_date)""",
        "predictions_query": """
select model_id,
      as_of_date,
      entity_id,
      score,
      label_value,
      coalesce(rank_abs_no_ties, row_number() over (partition by (model_id, as_of_date) order by score desc)) as rank_abs,
      coalesce(rank_pct_no_ties*100, ntile(100) over (partition by (model_id, as_of_date) order by score desc)) as rank_pct
  from test_results.predictions
  JOIN models_dates_join_query USING(model_id, as_of_date)
  where model_id IN (select model_id from models_list_query)
  AND as_of_date in (select as_of_date from as_of_dates_query)""",
    })


@pytest.fixture(scope="module")
def sample_timechop_splits():
    return [
        {
            "feature_start_time": "2010-01-01T00:00:00",
            "feature_end_time": "2014-01-01T00:00:00",
            "label_start_time": "2011-01-01T00:00:00",
            "label_end_time": "2014-01-01T00:00:00",
            "train_matrix": {
                "first_as_of_time": "2011-06-01T00:00:00",
                "last_as_of_time": "2011-12-01T00:00:00",
                "matrix_info_end_time": "2012-06-01T00:00:00",
                "as_of_times": [
                    "2011-06-01T00:00:00",
                    "2011-07-01T00:00:00",
                    "2011-08-01T00:00:00",
                    "2011-09-01T00:00:00",
                    "2011-10-01T00:00:00",
                    "2011-11-01T00:00:00",
                    "2011-12-01T00:00:00"
                ],
                "training_label_timespan": "6months",
                "training_as_of_date_frequency": "1month",
                "max_training_history": "6months"
            },
            "test_matrices": [
                {
                    "first_as_of_time": "2012-06-01T00:00:00",
                    "last_as_of_time": "2012-06-01T00:00:00",
                    "matrix_info_end_time": "2012-12-01T00:00:00",
                    "as_of_times": [
                        "2012-06-01T00:00:00"
                    ],
                    "test_label_timespan": "6months",
                    "test_as_of_date_frequency": "3months",
                    "test_duration": "1months"
                }
            ],
            "train_uuid": "40de3a41a7b210c6a525adeb74fafb22",
            "test_uuids": [
                "6c41a75c5270ed036370ca2344371150"
            ]
        },
        {
            "feature_start_time": "2010-01-01T00:00:00",
            "feature_end_time": "2014-01-01T00:00:00",
            "label_start_time": "2011-01-01T00:00:00",
            "label_end_time": "2014-01-01T00:00:00",
            "train_matrix": {
                "first_as_of_time": "2012-06-01T00:00:00",
                "last_as_of_time": "2012-12-01T00:00:00",
                "matrix_info_end_time": "2013-06-01T00:00:00",
                "as_of_times": [
                    "2012-06-01T00:00:00",
                    "2012-07-01T00:00:00",
                    "2012-08-01T00:00:00",
                    "2012-09-01T00:00:00",
                    "2012-10-01T00:00:00",
                    "2012-11-01T00:00:00",
                    "2012-12-01T00:00:00"
                ],
                "training_label_timespan": "6months",
                "training_as_of_date_frequency": "1month",
                "max_training_history": "6months"
            },
            "test_matrices": [
                {
                    "first_as_of_time": "2013-06-01T00:00:00",
                    "last_as_of_time": "2013-06-01T00:00:00",
                    "matrix_info_end_time": "2013-12-01T00:00:00",
                    "as_of_times": [
                        "2013-06-01T00:00:00"
                    ],
                    "test_label_timespan": "6months",
                    "test_as_of_date_frequency": "3months",
                    "test_duration": "1months"
                }
            ],
            "train_uuid": "95f998f70d5be1cf3d2ec833cd9db079",
            "test_uuids": [
                "8fd8be5c0b8b2e5b06a233b960769ccf"
            ]
        }
    ]


@pytest.fixture(scope="module")
def sample_grid_config():
    return {
        'sklearn.tree.DecisionTreeClassifier': {
            'max_depth': [2,10],
            'min_samples_split': [2],
        },
        'sklearn.ensemble.ExtraTreesClassifier': {
            'n_jobs': [-1],
            'n_estimators': [10],
            'criterion': ['gini'],
            'max_depth': [1],
            'max_features': ['sqrt'],
            'min_samples_split': [2,5]
        },
        'sklearn.ensemble.GradientBoostingClassifier': {
            'loss': ['deviance', 'exponential']
        },
        'triage.component.catwalk.estimators.classifiers.ScaledLogisticRegression': {
            'penalty': ['l1', 'l2'],
            'C': [0.01, 1]
        },
        'triage.component.catwalk.baselines.rankers.PercentileRankOneFeature': {
            'feature': ['feature_one', 'feature_two'],
            'descend': [True]
        },
        'triage.component.catwalk.baselines.thresholders.SimpleThresholder': {
            'rules': [['feature_one > 3', 'feature_two <= 5']],
            'logical_operator': ['and']
        }
    }



import io
import json

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import pandas as pd

from sqlalchemy.orm import sessionmaker

from triage.component.results_schema import Matrix
from triage.database_reflection import table_has_data
from triage.tracking import built_matrix, skipped_matrix, errored_matrix
from triage.util.pandas import downcast_matrix


class BuilderBase:
    def __init__(
        self,
        db_config,
        matrix_storage_engine,
        engine,
        experiment_hash,
        replace=True,
        include_missing_labels_in_train_as=None,
        run_id=None,
    ):
        self.db_config = db_config
        self.matrix_storage_engine = matrix_storage_engine
        self.db_engine = engine
        self.experiment_hash = experiment_hash
        self.replace = replace
        self.include_missing_labels_in_train_as = include_missing_labels_in_train_as
        self.run_id = run_id

    @property
    def sessionmaker(self):
        return sessionmaker(bind=self.db_engine)

    def validate(self):
        for expected_db_config_val in [
            "features_schema_name",
            "cohort_table_name",
            "labels_schema_name",
            "labels_table_name",
        ]:
            if expected_db_config_val not in self.db_config:
                raise ValueError(
                    "{} needed in db_config".format(expected_db_config_val)
                )

    def build_all_matrices(self, build_tasks):
        logger.info(f"Building {len(build_tasks.keys())} matrices")

        for i, (matrix_uuid, task_arguments) in enumerate(build_tasks.items(), start=1):
            logger.info(
                f"Building matrix {matrix_uuid} [{i}/{len(build_tasks.keys())}]"
            )
            self.build_matrix(**task_arguments)
            logger.success(f"Matrix {matrix_uuid} built")

    def _outer_join_query(
        self,
        right_table_name,
        right_column_selections,
        entity_date_table_name,
        additional_conditions="",
    ):
        """ Given a (features or labels) table, a list of times, columns to
        select, and (optionally) a set of join conditions, perform an outer
        join to the entity date table.

        :param right_table_name: the name of the right (feature/label) table
        :param right_column_selections: formatted text for the columns to select
        :param entity_date_table_name: name of table containing all valid entity ids and dates
        :param additional_conditions: formatted text for additional join
                                      conditions
        :type right_table_name: str
        :type right_column_selections: str
        :type entity_date_table_name: str
        :type additional_conditions: str

        :return: postgresql query for the outer join to the entity-dates table
        :rtype: str
        """

        # put everything into the query
        query = f"""
            SELECT ed.entity_id,
                   ed.as_of_date{"".join(right_column_selections)}
            FROM {entity_date_table_name} ed
            LEFT OUTER JOIN {right_table_name} r
            ON ed.entity_id = r.entity_id AND
               ed.as_of_date = r.as_of_date
               {additional_conditions}
            ORDER BY ed.entity_id,
                     ed.as_of_date
        """
        return query

    def make_entity_date_table(
        self,
        as_of_times,
        label_name,
        label_type,
        state,
        matrix_type,
        matrix_uuid,
        label_timespan,
    ):
        """ Make a table containing the entity_ids and as_of_dates required for
        the current matrix.

        :param as_of_times: the times to be used for the current matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param state: the entity state to be used in the matrix
        :param matrix_type: the type (train/test) of matrix
        :param matrix_uuid: a unique id for the matrix
        :param label_timespan: the time timespan that labels in matrix will include
        :type as_of_times: list
        :type label_name: str
        :type label_type: str
        :type state: str
        :type matrix_type: str
        :type matrix_uuid: str
        :type label_timespan: str

        :return: table name
        :rtype: str
        """

        as_of_time_strings = [str(as_of_time) for as_of_time in as_of_times]
        if matrix_type == "test" or self.include_missing_labels_in_train_as is not None:
            indices_query = self._all_valid_entity_dates_query(
                as_of_time_strings=as_of_time_strings, state=state
            )
        elif matrix_type == "train":
            indices_query = self._all_labeled_entity_dates_query(
                as_of_time_strings=as_of_time_strings,
                state=state,
                label_name=label_name,
                label_type=label_type,
                label_timespan=label_timespan,
            )
        else:
            raise ValueError(f"Unknown matrix type passed: {matrix_type}")

        table_name = "_".join([matrix_uuid, "matrix_entity_date"])
        query = f"""
            DROP TABLE IF EXISTS {self.db_config["features_schema_name"]}."{table_name}";
            CREATE TABLE {self.db_config["features_schema_name"]}."{table_name}"
            AS ({indices_query})
        """
        logger.debug(
            f"Creating matrix-specific entity-date table for matrix {matrix_uuid} ",
        )
        logger.spam(f"with query {query}")
        self.db_engine.execute(query)

        return table_name

    def _all_labeled_entity_dates_query(
        self, as_of_time_strings, state, label_name, label_type, label_timespan
    ):
        query = f"""
            SELECT entity_id, as_of_date
            FROM {self.db_config["cohort_table_name"]}
            JOIN {self.db_config["labels_schema_name"]}.{self.db_config["labels_table_name"]} using (entity_id, as_of_date)
            WHERE {state}
            AND as_of_date IN (SELECT (UNNEST (ARRAY{as_of_time_strings}::timestamp[])))
            AND label_name = '{label_name}'
            AND label_type = '{label_type}'
            AND label_timespan = '{label_timespan}'
            AND label is not null
            ORDER BY entity_id, as_of_date
        """
        return query

    def _all_valid_entity_dates_query(self, state, as_of_time_strings):
        query = f"""
            SELECT entity_id, as_of_date
            FROM {self.db_config["cohort_table_name"]}
            WHERE {state}
            AND as_of_date IN (SELECT (UNNEST (ARRAY{as_of_time_strings}::timestamp[])))
            ORDER BY entity_id, as_of_date
        """
        if not table_has_data(
            self.db_config["cohort_table_name"], self.db_engine
        ):
            raise ValueError("Required cohort table does not exist")
        return query


class MatrixBuilder(BuilderBase):
    def build_matrix(
        self,
        as_of_times,
        label_name,
        label_type,
        feature_dictionary,
        matrix_metadata,
        matrix_uuid,
        matrix_type,
    ):
        """ Write a design matrix to disk with the specified paramters.

        :param as_of_times: datetimes to be included in the matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param feature_dictionary: a dictionary of feature tables and features
                                   to be included in the matrix
        :param matrix_metadata: a dictionary of metadata about the matrix
        :param matrix_uuid: a unique id for the matrix
        :param matrix_type: the type (train/test) of matrix
        :type as_of_times: list
        :type label_name: str
        :type label_type: str
        :type feature_dictionary: dict
        :type matrix_metadata: dict
        :type matrix_uuid: str
        :type matrix_type: str

        :return: none
        :rtype: none
        """
        logger.spam(f"popped matrix {matrix_uuid} build off the queue")
        if not table_has_data(
            self.db_config["cohort_table_name"], self.db_engine
        ):
            logger.warning("cohort table is not populated, cannot build matrix")
            if self.run_id:
                errored_matrix(self.run_id, self.db_engine)
            return
        if not table_has_data(
                f"{self.db_config['labels_schema_name']}.{self.db_config['labels_table_name']}",
                self.db_engine,
        ):
            logger.warning("labels table is not populated, cannot build matrix")
            if self.run_id:
                errored_matrix(self.run_id, self.db_engine)
            return

        matrix_store = self.matrix_storage_engine.get_store(matrix_uuid)
        if not self.replace and matrix_store.exists:
            logger.notice(f"Skipping {matrix_uuid} because matrix already exists")
            if self.run_id:
                skipped_matrix(self.run_id, self.db_engine)
            return

        logger.debug(
            f'Storing matrix {matrix_metadata["matrix_id"]} in {matrix_store.matrix_base_store.path}'
        )
        # make the entity time table and query the labels and features tables
        logger.debug(f"Making entity date table for matrix {matrix_uuid}")
        try:
            entity_date_table_name = self.make_entity_date_table(
                as_of_times,
                label_name,
                label_type,
                matrix_metadata["state"],
                matrix_type,
                matrix_uuid,
                matrix_metadata["label_timespan"],
            )
        except ValueError as e:
            logger.exception(
                "Not able to build entity-date table,  will not build matrix",
            )
            if self.run_id:
                errored_matrix(self.run_id, self.db_engine)
            return
        logger.spam(
            f"Extracting feature group data from database into file  for matrix {matrix_uuid}"
        )
        dataframes = self.load_features_data(
            as_of_times, feature_dictionary, entity_date_table_name, matrix_uuid
        )
        logger.debug(f"Feature data extracted for matrix {matrix_uuid}")
        logger.spam(
            "Extracting label data from database into file for matrix {matrix_uuid}",
        )
        labels_df = self.load_labels_data(
            label_name,
            label_type,
            entity_date_table_name,
            matrix_uuid,
            matrix_metadata["label_timespan"],
        )
        dataframes.insert(0, labels_df)

        logger.debug(f"Label data extracted for matrix {matrix_uuid}")
        # stitch together the csvs
        logger.spam(f"Merging feature files for matrix {matrix_uuid}")
        output = self.merge_feature_csvs(dataframes, matrix_uuid)
        logger.debug(f"Features data merged for matrix {matrix_uuid}")

        matrix_store.metadata = matrix_metadata
        # store the matrix
        labels = output.pop(matrix_store.label_column_name)
        matrix_store.matrix_label_tuple = output, labels
        matrix_store.save()
        logger.info(f"Matrix {matrix_uuid} saved in {matrix_store.matrix_base_store.path}")
        # If completely archived, save its information to matrices table
        # At this point, existence of matrix already tested, so no need to delete from db
        if matrix_type == "train":
            lookback = matrix_metadata["max_training_history"]
        else:
            lookback = matrix_metadata["test_duration"]

        matrix = Matrix(
            matrix_id=matrix_metadata["matrix_id"],
            matrix_uuid=matrix_uuid,
            matrix_type=matrix_type,
            labeling_window=matrix_metadata["label_timespan"],
            num_observations=len(output),
            lookback_duration=lookback,
            feature_start_time=matrix_metadata["feature_start_time"],
            feature_dictionary=feature_dictionary,
            matrix_metadata=matrix_metadata,
            built_by_experiment=self.experiment_hash
        )
        session = self.sessionmaker()
        session.merge(matrix)
        session.commit()
        session.close()
        if self.run_id:
            built_matrix(self.run_id, self.db_engine)


    def load_labels_data(
        self,
        label_name,
        label_type,
        entity_date_table_name,
        matrix_uuid,
        label_timespan,
    ):
        """ Query the labels table and write the data to disk in csv format.

        :param as_of_times: the times to be used for the current matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param entity_date_table_name: the name of the entity date table
        :param matrix_uuid: a unique id for the matrix
        :param label_timespan: the time timespan that labels in matrix will include
        :type label_name: str
        :type label_type: str
        :type entity_date_table_name: str
        :type matrix_uuid: str
        :type label_timespan: str

        :return: name of csv containing labels
        :rtype: str
        """
        if self.include_missing_labels_in_train_as is None:
            label_predicate = "r.label"
        elif self.include_missing_labels_in_train_as is False:
            label_predicate = "coalesce(r.label, 0)"
        elif self.include_missing_labels_in_train_as is True:
            label_predicate = "coalesce(r.label, 1)"
        else:
            raise ValueError(
                f'incorrect value "{self.include_missing_labels_in_train_as}" for include_missing_labels_in_train_as'
            )

        labels_query = self._outer_join_query(
            right_table_name=f'{self.db_config["labels_schema_name"]}.{self.db_config["labels_table_name"]}',
            entity_date_table_name=f'"{self.db_config["features_schema_name"]}"."{entity_date_table_name}"',
            right_column_selections=f", {label_predicate} as {label_name}",
            additional_conditions=f"""AND
                r.label_name = '{label_name}' AND
                r.label_type = '{label_type}' AND
                r.label_timespan = '{label_timespan}'
            """
        )

        return self.query_to_df(labels_query)

    def load_features_data(
        self, as_of_times, feature_dictionary, entity_date_table_name, matrix_uuid
    ):
        """ Loop over tables in features schema, writing the data from each to a
        csv. Return the full list of feature csv names and the list of all
        features.

        :param as_of_times: the times to be included in the matrix
        :param feature_dictionary: a dictionary of feature tables and features
            to be included in the matrix
        :param entity_date_table_name: the name of the entity date table
            for the matrix
        :param matrix_uuid: a human-readable id for the matrix
        :type as_of_times: list
        :type feature_dictionary: dict
        :type entity_date_table_name: str
        :type matrix_uuid: str

        :return: list of csvs containing feature data
        :rtype: tuple
        """
        # iterate! for each table, make query, write csv, save feature & file names
        feature_dfs = []
        for feature_table_name, feature_names in feature_dictionary.items():
            logger.spam(f"Retrieving feature data from {feature_table_name}")
            features_query = self._outer_join_query(
                right_table_name=f'{self.db_config["features_schema_name"]}.{feature_table_name}',
                entity_date_table_name=f'{self.db_config["features_schema_name"]}."{entity_date_table_name}"',
                # collate imputation shouldn't leave any nulls and we double-check
                # the imputed table in FeatureGenerator.create_all_tables() but as
                # a final check, raise a divide by zero error on export if the
                # database encounters any during the outer join
                right_column_selections=[', "{0}"'.format(fn) for fn in feature_names],
            )
            feature_dfs.append(self.query_to_df(features_query))

        return feature_dfs

    def query_to_df(self, query_string, header="HEADER"):
        """ Given a query, write the requested data to csv.

        :param query_string: query to send
        :param file_name: name to save the file as
        :header: text to include in query indicating if a header should be saved
                 in output
        :type query_string: str
        :type file_name: str
        :type header: str

        :return: none
        :rtype: none
        """
        logger.spam(f"Copying to CSV query {query_string}")
        copy_sql = f"COPY ({query_string}) TO STDOUT WITH CSV {header}"
        conn = self.db_engine.raw_connection()
        cur = conn.cursor()
        out = io.StringIO()
        cur.copy_expert(copy_sql, out)
        out.seek(0)
        df = pd.read_csv(out, parse_dates=["as_of_date"])
        df.set_index(["entity_id", "as_of_date"], inplace=True)
        return downcast_matrix(df)

    def merge_feature_csvs(self, dataframes, matrix_uuid):
        """Horizontally merge a list of feature CSVs
        Assumptions:
        - The first and second columns of each CSV are
          the entity_id and date
        - That the CSVs have the same list of entity_id/date combinations
          in the same order.
        - The first CSV is expected to be labels, and only have
          entity_id, date, and label.
        - All other CSVs do not have any labels (all non entity_id/date columns
          will be treated as features)
        - The label will be in the *last* column of the merged CSV

        :param source_filenames: the filenames of each feature csv
        :param out_filename: the desired filename of the merged csv
        :type source_filenames: list
        :type out_filename: str

        :return: none
        :rtype: none

        :raises: ValueError if the first two columns in every CSV don't match
        """

        for i, df in enumerate(dataframes):
            if df.index.names != ["entity_id", "as_of_date"]:
                raise ValueError(
                    f"index must be entity_id and as_of_date, value was {df.index}"
                )
            # check for any nulls. the labels, understood to be the first file,
            # can have nulls but no features should. therefore, skip the first dataframe
            if i > 0:
                columns_with_nulls = [
                    column for column in df.columns if df[column].isnull().values.any()
                ]
                if len(columns_with_nulls) > 0:
                    raise ValueError(
                        "Imputation failed for the following features: {columns_with_nulls}"
                    )
            i += 1

        big_df = dataframes[1].join(dataframes[2:] + [dataframes[0]])
        return big_df



import itertools

from triage.component.architect.feature_group_mixer import FeatureGroupMixer
from triage.component.architect.feature_group_creator import FeatureGroup

import pytest


@pytest.fixture
def english_numbers():
    return FeatureGroup(
        name="english_numbers",
        features_by_table={"one": ["two", "three"], "four": ["five", "six"]},
    )


@pytest.fixture
def letters():
    return FeatureGroup(
        name="letters", features_by_table={"a": ["b", "c"], "d": ["e", "f"]}
    )


@pytest.fixture
def german_numbers():
    return FeatureGroup(
        name="german_numbers",
        features_by_table={"eins": ["zwei", "drei"], "vier": ["funf", "sechs"]},
    )


def test_feature_group_mixer_leave_one_out(english_numbers, letters, german_numbers):
    feature_groups = [english_numbers, letters, german_numbers]

    result = FeatureGroupMixer(["leave-one-out"]).generate(feature_groups)
    expected = [
        dict(itertools.chain(letters.items(), german_numbers.items())),
        dict(itertools.chain(english_numbers.items(), german_numbers.items())),
        dict(itertools.chain(english_numbers.items(), letters.items())),
    ]
    assert result == expected
    assert [g.names for g in result] == [
        ["letters", "german_numbers"],
        ["english_numbers", "german_numbers"],
        ["english_numbers", "letters"],
    ]


def test_feature_group_mixer_leave_one_in(english_numbers, letters, german_numbers):
    feature_groups = [english_numbers, letters, german_numbers]

    result = FeatureGroupMixer(["leave-one-in"]).generate(feature_groups)
    expected = [english_numbers, letters, german_numbers]
    assert result == expected
    assert [g.names for g in result] == [
        ["english_numbers"],
        ["letters"],
        ["german_numbers"],
    ]


def test_feature_group_mixer_all_combinations(english_numbers, letters,
                                              german_numbers):
    feature_groups = [english_numbers, letters, german_numbers]

    result = FeatureGroupMixer(['all-combinations']).generate(feature_groups)
    expected = [
        dict(itertools.chain(english_numbers.items())),
        dict(itertools.chain(letters.items())),
        dict(itertools.chain(german_numbers.items())),
        dict(itertools.chain(english_numbers.items(), letters.items())),
        dict(itertools.chain(english_numbers.items(), german_numbers.items())),
        dict(itertools.chain(letters.items(), german_numbers.items())),
        dict(itertools.chain(english_numbers.items(), letters.items(),
                             german_numbers.items()))
    ]
    assert result == expected
    assert [g.names for g in result] == [
        ['english_numbers'],
        ['letters'],
        ['german_numbers'],
        ['english_numbers', 'letters'],
        ['english_numbers', 'german_numbers'],
        ['letters', 'german_numbers'],
        ['english_numbers', 'letters', 'german_numbers']
    ]

def test_feature_group_mixer_all(english_numbers, letters, german_numbers):
    feature_groups = [english_numbers, letters, german_numbers]

    result = FeatureGroupMixer(["all"]).generate(feature_groups)
    expected = [
        dict(
            itertools.chain(
                english_numbers.items(), letters.items(), german_numbers.items()
            )
        )
    ]
    assert result == expected
    assert result[0].names == ["english_numbers", "letters", "german_numbers"]



import copy
import itertools

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from triage.component.catwalk.utils import filename_friendly_hash
from . import utils, entity_date_table_generators


class Planner:
    def __init__(
        self,
        feature_start_time,
        label_names,
        label_types,
        cohort_names,
        user_metadata,
    ):
        self.feature_start_time = (
            feature_start_time
        )  # earliest time included in features
        self.label_names = label_names
        self.label_types = label_types
        self.cohort_names = cohort_names
        self.user_metadata = user_metadata

    def _generate_build_task(
        self, matrix_metadata, matrix_uuid, train_matrix, feature_dictionary
    ):
        return {
            "as_of_times": train_matrix["as_of_times"],
            "label_name": matrix_metadata["label_name"],
            "label_type": matrix_metadata["label_type"],
            "feature_dictionary": feature_dictionary,
            "matrix_uuid": matrix_uuid,
            "matrix_metadata": matrix_metadata,
            "matrix_type": matrix_metadata["matrix_type"],
        }

    def _make_metadata(
        self,
        matrix_definition,
        feature_dictionary,
        label_name,
        label_type,
        cohort_name,
        matrix_type,
    ):
        """ Generate dictionary of matrix metadata.

        :param matrix_definition: temporal definition of matrix
        :param feature dictionary: feature tables and the columns within them to use as features
        :param label_name: name of label column
        :param label_type: type of label
        :param cohort_name: the cohort name to be included in the matrix
        :param matrix_type: type (train/test) of matrix
        :type matrix_definition: dict
        :type feature dictionary: dict
        :type label_name: str
        :type label_type: str
        :type cohort_name: str
        :type matrix_type: str

        :return: metadata needed for matrix identification and modeling
        :rtype: dict
        """

        # make a human-readable label for this matrix
        matrix_id = "_".join(
            [
                label_name,
                label_type,
                str(matrix_definition["first_as_of_time"]),
                str(matrix_definition["matrix_info_end_time"]),
            ]
        )
        matrix_metadata = {
            # temporal information
            "feature_start_time": self.feature_start_time,
            "end_time": matrix_definition["matrix_info_end_time"],
            "as_of_date_frequency": matrix_definition.get(
                "training_as_of_date_frequency",
                matrix_definition.get("test_as_of_date_frequency"),
            ),
            # columns
            "indices": ["entity_id", "as_of_date"],
            "feature_names": utils.feature_list(feature_dictionary),
            "feature_groups": feature_dictionary.names,
            "label_name": label_name,
            # other information
            "label_type": label_type,
            "label_timespan": matrix_definition.get(
                "test_label_timespan",
                matrix_definition.get("training_label_timespan", "0 days"),
            ),
            "state": entity_date_table_generators.DEFAULT_ACTIVE_STATE,
            "cohort_name": cohort_name,
            "matrix_id": matrix_id,
            "matrix_type": matrix_type,
        }
        matrix_metadata.update(matrix_definition)
        matrix_metadata.update(self.user_metadata)

        return matrix_metadata

    def generate_plans(self, matrix_set_definitions, feature_dictionaries):
        """Create build tasks and update the matrix definitions with UUIDs

        :param matrix_set_definitions: the temporal information needed to generate each matrix
        :param feature_dictionaries: combinations of features to include in matrices
        :type matrix_set_definitions: list
        :type feature_dictionaries: list

        :return: matrix set definitions (updated with matrix uuids) and build tasks
        :rtype: tuple (list, dict)
        """
        updated_definitions = []
        build_tasks = dict()
        for matrix_set in matrix_set_definitions:
            logger.debug("Making plans for matrix set %s", matrix_set)
            logger.debug(
                "Iterating over %s label names, %s label_types, %s cohort_names, "
                "%s feature dictionaries",
                len(self.label_names),
                len(self.label_types),
                len(self.cohort_names),
                len(feature_dictionaries),
            )
            train_matrix = matrix_set["train_matrix"]
            for (
                label_name,
                label_type,
                cohort_name,
                feature_dictionary,
            ) in itertools.product(
                self.label_names, self.label_types, self.cohort_names, feature_dictionaries
            ):
                matrix_set_clone = copy.deepcopy(matrix_set)
                # get a uuid
                train_metadata = self._make_metadata(
                    train_matrix,
                    feature_dictionary,
                    label_name,
                    label_type,
                    cohort_name,
                    "train",
                )
                train_uuid = filename_friendly_hash(train_metadata)
                logger.debug(
                    "Matrix UUID %s found for train metadata %s",
                    train_uuid,
                    train_metadata,
                )
                if train_uuid not in build_tasks:
                    build_tasks[train_uuid] = self._generate_build_task(
                        train_metadata, train_uuid, train_matrix, feature_dictionary
                    )
                    logger.debug(
                        "Train uuid %s not found in build tasks yet, " "so added",
                        train_uuid,
                    )
                else:
                    logger.debug(
                        "Train uuid %s already found in build tasks", train_uuid
                    )
                matrix_set_clone["train_uuid"] = train_uuid

                test_uuids = []
                for test_matrix in matrix_set_clone["test_matrices"]:
                    test_metadata = self._make_metadata(
                        test_matrix,
                        feature_dictionary,
                        label_name,
                        label_type,
                        cohort_name,
                        "test",
                    )
                    test_uuid = filename_friendly_hash(test_metadata)
                    logger.debug(
                        "Matrix UUID %s found for test metadata %s",
                        test_uuid,
                        test_metadata,
                    )
                    if test_uuid not in build_tasks:
                        build_tasks[test_uuid] = self._generate_build_task(
                            test_metadata, test_uuid, test_matrix, feature_dictionary
                        )
                        logger.debug(
                            "Test uuid %s not found in build tasks " "yet, so added",
                            test_uuid,
                        )
                    else:
                        logger.debug(
                            "Test uuid %s already found in build tasks", test_uuid
                        )

                    test_uuids.append(test_uuid)
                matrix_set_clone["test_uuids"] = test_uuids
                updated_definitions.append(matrix_set_clone)

        logger.debug(
            "Planner is finished generating matrix plans. "
            "%s matrix definitions and %s unique build tasks found",
            len(updated_definitions),
            len(build_tasks.keys()),
        )
        logger.debug("Associated all tasks with experiment in database")
        return updated_definitions, build_tasks
